////////////////////////////////////////////////////////////////////// 
// metal/Main.cpp 
// (c) 2000-2008 Goncalo Abecasis
// 
// This file is distributed as part of the Goncalo source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile Goncalo.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Thursday January 10, 2008
// 
 
#include "StringArray.h"
#include "StringHash.h"
#include "MathVector.h"
#include "MathStats.h"

#include <math.h>

StringIntHash markerLookup;

Vector statistics;
Vector weights;
Vector pvalues;

StringArray allele1;
StringArray allele2;
String      original_flipped;

StringArray filenames;
StringArray directions;

double weight = 1.0;
double minweight = 1.0;

String markerLabel  = "MARKER";
String weightLabel  = "N";
String pvalueLabel  = "PVALUE";
String effectLabel  = "EFFECT";
String firstAllele  = "ALLELE1";
String secondAllele = "ALLELE2";

String separators  = " \t";

int    maxWarnings = 20;

void ClearAll()
   {
   printf("## Clearing all stored statistics ...\n");
   markerLookup.Clear();
   statistics.Dimension(0);
   weights.Dimension(0);
   pvalues.Dimension(0);
   allele1.Dimension(0);
   allele2.Dimension(0);
   original_flipped.Clear();
   filenames.Clear();
   directions.Clear();
   }

void NumbersToLetters(String & al)
   {
   if (al == "1") al = "a"; else
   if (al == "2") al = "c"; else
   if (al == "3") al = "g"; else
   if (al == "4") al = "t";
   }

void FlipAllele(String & al)
   {
   if (al == "A" || al == "a") al = "t"; else
   if (al == "C" || al == "c") al = "g"; else
   if (al == "G" || al == "g") al = "c"; else
   if (al == "T" || al == "t") al = "a";
   }

bool FlipAlleles(String & al1, String & al2, double & effect)
   {
   al1.ToLower();
   al2.ToLower();

   if (al1 > al2)
      { al1.Swap(al2); effect *= -1.0; }

   if (al1 == "a" || al1 == "c" && al2 == "g")
      return false;

   FlipAllele(al1);
   FlipAllele(al2);

   if (al1 > al2)
      { al1.Swap(al2); effect *= -1.0; }

   return true;
   }

bool GuessSecondAllele(int marker, String & al1, String & al2)
   {
   al1.ToLower();

   FlipAllele(al2 = al1);

   if (allele1[marker] == al1 && allele2[marker] == al2 ||
       allele1[marker] == al2 && allele2[marker] == al1)
      return false;

   if (allele1[marker] == al1)
      {
      al2 = allele2[marker];
      return true;
      }
   else if (allele2[marker] == al1)
      {
      al2 = allele1[marker];
      return true;
      }
   else if (allele1[marker] == al2)
      {
      FlipAllele(al2 = allele2[marker]);
      return true;
      }
   else if (allele2[marker] == al2)
      {
      FlipAllele(al2 = allele1[marker]);
      return true;
      }

   return false;
   }

int    outfileround = 1;
String outfile("METAANALYSIS%d.TBL");

void Analyze()
   {
   String filename;
   filename.printf(outfile, outfileround++);

   printf("## Executing meta-analysis ...\n");
   printf("## Complete results will be stored in file '%s'\n", (const char *) filename);
   printf("## Column descriptions will be stored in file '%s.info'\n", (const char *) filename);

   FILE * f = fopen(filename, "wt");

   fprintf(f, "Marker\tAllele1\tAllele2\tWeight\tZscore\tP-value\tDirection\n");

   double smallp = 1.0;
   String smallm("nothing tested");

   String al1, al2;

   String direction;

   int count = 0, skipped = 0;
   for (int i = 0; i < markerLookup.Capacity(); i++)
      if (markerLookup.SlotInUse(i))
         {
         int marker = markerLookup.Integer(i);

         if (weights[marker] == 0.0) continue;

         if (weights[marker] < minweight)
            {
            skipped++;
            continue;
            }

         double statistic = statistics[marker] / sqrt(weights[marker]);
         double pvalue = 2.0 * ndist(fabs(statistic), true);

         al1 = allele1[marker];
         al2 = allele2[marker];

         if (original_flipped[marker] == 'Y')
            FlipAllele(al1), FlipAllele(al2);

         direction.Clear();
         for (int j = 0; j < filenames.Length(); j++)
            direction += marker < directions[j].Length() ? directions[j][marker] : '?';

         fprintf(f, "%s\t%s\t%s\t%.2f\t%.3f\t%.4g\t%s\n",
                  (const char *) markerLookup[i],
                  (const char *) al1,
                  (const char *) al2,
                  weights[marker],
                  statistic,
                  pvalue,
                  (const char *) direction);
         count++;

         if (pvalue < smallp)
            {
            smallp = pvalue;
            smallm = markerLookup[i];
            }
         }

   printf("## Completed meta-analysis for %d markers!\n", count);

   if (smallp < 1.0)
      printf("## Smallest p-value is %.4g at marker '%s'\n", smallp, (const char *) smallm);

   if (skipped)
      printf("## A total of %d markers with less than MINWEIGHT = %g were skipped\n", skipped, minweight);

   fclose(f);

   f = fopen(filename + ".info", "wt");

   fprintf(f, "# This file contains a short description of the columns in the\n"
              "# meta-analysis summary file, named '%s'\n\n"
              "# Marker    - this is the marker name\n"
              "# Allele1   - the first allele for this marker in the first file where it occurs\n"
              "# Allele2   - the second allele for this marker in the first file where it occurs\n"
              "# Weight    - the sum of the individual study weights (typically, N) for this marker\n"
              "# Z-score   - the combined z-statistic for this marker\n"
              "# Direction - summary of effect direction for each study, with one '+' or '-' per study\n\n"
              "# Input for this meta-analysis was stored in the files:\n",
              (const char *) filename);

   for (int i = 0; i < filenames.Length(); i++)
      fprintf(f, "# --> Input File %d : %s\n", i + 1, (const char *) filenames[i]);
   fprintf(f, "\n");

   fclose(f);
   }

bool flip = false;

void ProcessFile(String & filename)
   {
   IFILE f = ifopen(filename, "rb");

   if (f == NULL)
      {
      printf("## Failed to open file '%s'\n", (const char *) filename);
      return;
      }

   printf("## Processing file '%s'\n", (const char *) filename);

   String input;
   StringArray tokens;

   tokens.ReplaceTokens(input.ReadLine(f), separators);

   if (tokens.Length() == 0)
      {
      printf("## ERROR: Header line is empty\n");
      ifclose(f);
      return;
      }

   int markerColumn = tokens.SlowFind(markerLabel);
   if (markerColumn < 0)
      {
      printf("## ERROR: No '%s' column found\n", (const char *) markerLabel);
      ifclose(f);
      return;
      }

   int pvalueColumn = tokens.SlowFind(pvalueLabel);
   if (pvalueColumn < 0)
      {
      printf("## ERROR: No '%s' column found\n", (const char *) pvalueLabel);
      ifclose(f);
      return;
      }

   filenames.Push(filename);

   int effectColumn = tokens.SlowFind(effectLabel);
   if (effectColumn < 0)
      printf("## WARNING: No '%s' column found -- assuming all effects are positive!\n", (const char *) effectLabel);

   int weightColumn = tokens.SlowFind(weightLabel);
   if (weightColumn < 0)
      printf("## WARNING: No '%s' column found -- using DEFAULTWEIGHT = %g\n", (const char *) weightLabel, weight);

   int firstColumn = tokens.SlowFind(firstAllele);
   int secondColumn = tokens.SlowFind(secondAllele);
   bool guessAlleles = false;

   if (firstColumn >= 0 && secondColumn < 0)
      {
      printf("## WARNING: Column '%s' not found, will try to guess second allele\n",
             (const char *) secondAllele);
      guessAlleles = true;
      }

   if (firstColumn < 0)
      printf("## WARNING: Column '%s', '%s' or both not found -- assuming all effects refer to same allele\n",
             (const char *) firstAllele, (const char *) secondAllele, firstColumn = 0);

   int minColumns = markerColumn + 1;
   if (weightColumn >= minColumns) minColumns = weightColumn + 1;
   if (effectColumn >= minColumns) minColumns = effectColumn + 1;
   if (pvalueColumn >= minColumns) minColumns = pvalueColumn + 1;
   if (firstColumn >= minColumns)  minColumns = firstColumn + 1;
   if (secondColumn >= minColumns) minColumns = secondColumn + 1;

   String direction;
   direction.Fill('?', allele1.Length());

   int invalid = 0;
   int badAlleles = 0;
   int badGuesses = 0;
   int duplicates = 0;
   int blindGuesses = 0;

   while (!ifeof(f))
      {
      if (separators.Length() != 1)
         tokens.ReplaceTokens(input.ReadLine(f).Trim(), separators);
      else
         tokens.ReplaceColumns(input.ReadLine(f), separators[0]);

      if (input[0] == '#') continue;
      if (tokens.Length() < minColumns) continue;

      int marker = markerLookup.Integer(tokens[markerColumn]);

      if (marker < 0 && guessAlleles)
         {
         if (++blindGuesses > maxWarnings) continue;
         printf("## WARNING: Can't guess alleles for '%s', marker not seen in previous files\n",
                (const char *) tokens[markerColumn]);
         continue;
         }

      if (marker < 0)
         {
         markerLookup.SetInteger(tokens[markerColumn], marker = markerLookup.Entries());
         statistics.Push(0);
         weights.Push(0);
         allele1.Push("");
         allele2.Push("");
         original_flipped += '?';
         direction += '?';
         }

      double p = tokens[pvalueColumn].AsDouble();

      if (p <= 0 || p > 1.0)
         {
         if (++invalid > maxWarnings) continue;
         printf("## WARNING: Invalid pvalue for marker %s, ignored\n", (const char *) tokens[markerColumn]);
         continue;
         }

      double z = -ninv(tokens[pvalueColumn].AsDouble() * 0.5);
      double w = weightColumn >= 0 ? tokens[weightColumn].AsDouble() : weight;

      if (effectColumn >= 0 && tokens[effectColumn][0] == '-')
         z *= -1;

      if (flip)
         z *= -1;

      if (firstColumn >= 0 && secondColumn >= 0)
         {
         NumbersToLetters(tokens[firstColumn]);
         NumbersToLetters(tokens[secondColumn]);
         bool flip = FlipAlleles(tokens[firstColumn], tokens[secondColumn], z);

         if (allele1[marker] == "")
            allele1[marker] = tokens[firstColumn],
            allele2[marker] = tokens[secondColumn],
            original_flipped[marker] = flip ? 'Y' : 'N';
         else
            if (tokens[firstColumn] != allele1[marker] ||
                tokens[secondColumn] != allele2[marker])
                {
                if (++badAlleles > maxWarnings) continue;
                printf("## WARNING: Bad alleles for marker '%s', expecting '%s/%s' found '%s/%s'\n",
                         (const char *) tokens[markerColumn],
                         (const char *) allele1[marker],
                         (const char *) allele2[marker],
                         (const char *) tokens[firstColumn],
                         (const char *) tokens[secondColumn]);
                continue;
                }
         }
      else if (guessAlleles)
         {
         tokens.Push("");
         NumbersToLetters(tokens[firstColumn]);
         if (!GuessSecondAllele(marker, tokens[firstColumn], tokens.Last()))
             {
             if (++badGuesses <= maxWarnings)
                printf("## WARNING: Bad allele or ambiguous strand for marker '%s', expecting '%s/%s' found '%s'\n",
                            (const char *) tokens[markerColumn],
                            (const char *) allele1[marker],
                            (const char *) allele2[marker],
                            (const char *) tokens[firstColumn]);
             continue;
             }
         FlipAlleles(tokens[firstColumn], tokens.Last(), z);
         }

      if (direction[marker] != '?')
         {
         if (++duplicates > maxWarnings) continue;
         printf("## WARNING: Marker '%s' duplicated in input, first instance used, others skipped\n",
                (const char *) tokens[markerColumn]);
         continue;
         }

      direction[marker] = z == 0.0 ? '0' : (z > 0.0 ? '+' : '-');
      statistics[marker] += sqrt(w) * z;
      weights[marker] += w;
      }

   if (invalid > maxWarnings)
      printf("## WARNING: Invalid p-values for %d other markers also ignored\n", invalid - maxWarnings);

   if (badGuesses > maxWarnings)
      printf("## WARNING: Failed to guess second allele for %d other markers\n", badGuesses - maxWarnings);

   if (duplicates > maxWarnings)
      printf("## WARNING: An additional %d rows with duplicate marker names were ignored\n", duplicates - maxWarnings);

   if (badAlleles > maxWarnings)
      printf("## WARNING: Allele names don't match previous occurences at %d additional markers\n", badAlleles - maxWarnings);

   if (blindGuesses > maxWarnings)
      printf("## WARNING: An additional %d markers whose alleles couldn't be guessed were skipped\n", blindGuesses - maxWarnings);

   directions.Push(direction);

   ifclose(f);
   }

int main(int argc, char ** argv)
   {
   String::caseSensitive = false;

   printf("MetaAnalysis Helper - (c) 2007 Goncalo Abecasis\n\n");

   printf("Available Commands:\n"
          "   WEIGHTLABEL      [LABEL]          (default = '%s')\n"
          "   MARKERLABEL      [LABEL]          (default = '%s')\n"
          "   PVALUELABEL      [LABEL]          (default = '%s')\n"
          "   EFFECTLABEL      [LABEL]          (default = '%s')\n"
          "   ALLELELABELS     [LABEL1 LABEL2]  (default = '%s','%s')\n"
          "   OUTFILE          [PREFIX SUFFIX]  (default = 'METAANALYSIS','.TBL')\n"
          "   DEFAULTWEIGHT    [NUMBER]         (default = 1.0)\n"
          "   MINWEIGHT        [NUMBER]         (default = 0.0)\n"
          "   MAXWARNINGS      [NUMBER]         (default = %d)\n"
          "   FLIP\n"
          "   SEPARATOR        [WHITESPACE|COMMAS|BOTH|TABS]\n"
          "                                     (default = WHITESPACE)\n"
          "   PROCESSFILE      [FILENAME]\n"
          "   ANALYZE\n"
          "   CLEAR\n\n",
          (const char *) weightLabel,
          (const char *) markerLabel,
          (const char *) pvalueLabel,
          (const char *) effectLabel,
          (const char *) firstAllele, (const char *) secondAllele,
          maxWarnings);

   String input;
   StringArray tokens;

   while (!feof(stdin))
      {
      input.ReadLine().Trim();
      tokens.ReplaceTokens(input);

      if (tokens.Length() == 0)
         continue;

      if (input[0] == '#') continue;

      if ((tokens[0].MatchesBeginningOf("ANALYZE") == 0 ||
          tokens[0].MatchesBeginningOf("ANALYSE") == 0) &&
          tokens[0].Length() > 1)
         {
         Analyze();
         continue;
         }

      if (tokens[0].MatchesBeginningOf("CLEAR") == 0)
         {
         ClearAll();
         continue;
         }

      if (tokens[0].MatchesBeginningOf("FLIP") == 0)
         {
         flip = !flip;
         printf("## All effects will %s flipped\n", flip ? "be" : "not be");
         continue;
         }

      if (tokens[0].MatchesBeginningOf("QUIT") == 0 ||
          tokens[0].MatchesBeginningOf("EXIT") == 0 && tokens[0].Length() > 1)
         break;

      if (tokens.Length() == 1)
         {
         printf("##  Error processing your request ...\n");
         continue;
         }

      if (tokens[0].MatchesBeginningOf("DEFAULTWEIGHT") == 0)
         {
         weight = tokens[1].AsDouble();
         printf("## Set default weight to %.2f ...\n", weight);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("WEIGHTLABEL") == 0)
         {
         weightLabel = tokens[1];
         printf("## Set weight header to %s ...\n", (const char *) weightLabel);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("MARKERLABEL") == 0 &&
          tokens[0].Length() > 2)
         {
         markerLabel = tokens[1];
         printf("## Set marker header to %s ...\n", (const char *) markerLabel);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("MAXWARNINGS") == 0 &&
          tokens[0].Length() > 2)
          {
          maxWarnings = tokens[1].AsInteger();
          printf("## Set the maximum number of warnings per file to %d ...\n", maxWarnings);
          continue;
          }

      if (tokens[0].MatchesBeginningOf("PVALUELABEL") == 0 && tokens[0].Length() > 1)
         {
         pvalueLabel = tokens[1];
         printf("## Set p-value header to %s ...\n", (const char *) pvalueLabel);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("EFFECTLABEL") == 0 && tokens[0].Length() > 1)
         {
         effectLabel = tokens[1];
         printf("## Set effect header to %s ...\n", (const char *) effectLabel);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("SEPARATOR") == 0)
         if (tokens[1].MatchesBeginningOf("WHITESPACE") == 0)
            {
            separators = " \t";
            printf("## Set column separator to WHITESPACE ...\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("COMMAS") == 0)
            {
            separators = ",";
            printf("## Set column separator to COMMAS ...\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("TABS") == 0)
            {
            separators = "\t";
            printf("## Set column separator to TAB ...\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("BOTH") == 0)
            {
            separators = " \t,";
            printf("## Set column separator to BOTH ...\n");
            continue;
            }

      if (tokens[0].MatchesBeginningOf("PROCESSFILE") == 0 && tokens[0].Length() > 1)
         {
         ProcessFile(tokens[1]);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("MINWEIGHT") == 0 &&
          tokens[0].Length() > 1)
          {
          minweight = tokens[1].AsDouble();
          printf("## Set minimum weight for meta-analysis to %.2f ...\n", minweight);
          continue;
          }

      if (tokens.Length() == 2)
         {
         printf("##  Error processing your request ...\n");
         continue;
         }

      if (tokens[0].MatchesBeginningOf("ALLELELABELS") == 0 &&
          tokens[0].Length() > 1)
         {
         firstAllele = tokens[1];
         secondAllele = tokens[2];
         printf("## Set allele headers to %s and %s ...\n",
               (const char *) firstAllele, (const char *) secondAllele);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("OUTFILE") == 0)
         {
         outfileround = 1;
         outfile = tokens[1] + "%d" + tokens[2];
         printf("## Set output file prefix and suffix to %s and %s ...\n",
                 (const char *) tokens[1], (const char *) tokens[2]);
         continue;
         }

      printf("## Command not recognized\n");
      }
   }
 
